#!/usr/bin/perl -w

#######################################################################################
#
# authorize_ssh_access v 1.20
# (c) 2000 - 2003 junidas GmbH
#
#######################################################################################


# Globale Definitionen

# GPG
my $gpg_executable = "/usr/local/bin/gpg";
my $good_signature_pattern = "^gpg: Good signature from .*|^gpg: Korrekte Unterschrift von .*";
my $tmpdir= "/tmp";


# Logging
my $log_file = "/var/log/activatessh";

# Die Umgebungsvariable, in der der MTA den Sender speichert
my $sender=$ENV{"SENDER"};
chomp $sender;

# Der Aufruf zum Versenden der E-Mail.
# Sollte fuer die meisten MTAs funktionieren.
my $sendmail = "/usr/lib/sendmail -oi -t";

# Die maximale Differenzzeit in Sekunden, die bei einer eingehenden Email akzeptiert wird.
my $time_diff = 600;

# ab hier sollten keine Modifikationen noetig sein

use strict;

use English;
use Fcntl ':flock'; # importiere LOCK_*-Konstanten
use Time::Local;
use File::Temp qw/ tempfile /;

File::Temp->safe_level( File::Temp::HIGH );


sub lockLOG();
sub unlockLOG();
sub write_ip($$);
sub time_difference_ok($$$$$);
sub send_mail($$$);

$SIG{PIPE} = 'IGNORE';

open(LOG, ">>$log_file") or die "Cannot open logfile: $log_file";


# Variablen fuer das Analysieren der Email
my $parse_content = 0;
my $ip_found = 0;
my $time_ok = 0;
my $msg_count = 0;
my $message = "";
my $ip_address = "";
my $line;


while ($line = <STDIN>)
{
	$line =~ s/=3D/=/g; # Gleichzeichen innerhalb der Signatur als Quoted Printable rueckuebersetzen
	$message .= $line;

	if ($line =~ /-----BEGIN PGP SIGNED MESSAGE-----/)
	{
		$msg_count++;
		$parse_content=1;
		next;
	}
	if ($line =~ /-----BEGIN PGP SIGNATURE-----/)
	{
		$parse_content=0;
		next;
	}

	if ($parse_content == 1)
	{
		# IP Format: xxx.xxx.xxx.xxx
		if ($line =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/ )
		{
			$ip_address = $1;
			$ip_found = 1;
			next;
		}
		# Time / Date Format: Stunde:Minute, Tag.Monat.Jahr
		if ($line =~ /(\d+):(\d+)[\s,]+(\d+)\.(\d+)\.(\d+)/ )
		{
			$time_ok = 1 if time_difference_ok($1, $2, $3, $4, $5);
		}
		# Date / Time Format: Tag.Monat.Jahr, Stunde:Minute
		if ($line =~ /(\d+)\.(\d+)\.(\d+)[\s,]+(\d+):(\d+)/ )
		{
			$time_ok = 1 if time_difference_ok($4, $5, $1, $2, $3);
		}
	}
}

# Log schreiben
lockLOG();
print LOG "$PID: new mail\n" or die "Cannot write to logfile: $log_file";
print LOG "$message\n" or die "Cannot write to logfile: $log_file";
print LOG "$PID: end of message\n" or die "Cannot write to logfile: $log_file";
unlockLOG();

# GPG zur Pruefung aufrufen
my $error;
my $tmpfilehandle;
my $tmpfilename;

# sicheres Tmp-File erzeugen
eval
{
	($tmpfilehandle, $tmpfilename) = tempfile(UNLINK => 1, DIR => $tmpdir);
	flock($tmpfilehandle, LOCK_EX);
};

if($@)
{
	$error = "$PID: Cannot create temporary file\n";
	lockLOG();
	print LOG $error or die "Cannot write to logfile: $log_file";
	unlockLOG();
	die $error;
}
unless (print $tmpfilehandle $message)
{
	$error = "$PID: Cannot write to temporary file $tmpfilename\n";
	lockLOG();
	print LOG $error or die "Cannot write to logfile: $log_file";
	unlockLOG();
	die $error;
}
unless (close ($tmpfilehandle))
{
	$error = "$PID: Cannot close temporary file $tmpfilename.\n";
	lockLOG();
	print LOG $error or die "Cannot write to logfile: $log_file";
	unlockLOG();
}

# Resultat von GPG pruefen
my $signature_valid = 0;

unless (open(CHECK, "$gpg_executable --verify <$tmpfilename 2>&1 |"))
{
	$error = "$PID: Cannot execute $gpg_executable\n";
	lockLOG();
	print LOG $error or die "Cannot write to logfile: $log_file";
	unlockLOG();
	die $error;
}
while ($line = <CHECK>)
{
	# check for signature verification
	$signature_valid = 1 if $line =~ /$good_signature_pattern/;
}
unless (close(CHECK))
{
	$error = "$PID: Cannot close channel to $gpg_executable.\n";
	$error .= "$PID: Most probably due to non-existent signature.\n";
	lockLOG();
	print LOG $error or die "Cannot write to logfile: $log_file";
	unlockLOG();
}

# Auswertung
my $log_entry;

if($signature_valid == 0 or $time_ok == 0 or $msg_count != 1)
{
	$log_entry = "$PID: *** rejected $ip_address (not authorized) ***\n";
	$log_entry .= "$PID: signature is invalid\n" if $signature_valid == 0;
	$log_entry .= "$PID: time sent in e-mail is not in allowed time interval\n" if $time_ok == 0;
	$log_entry .= "$PID: wrong number of PGP indicators\n" if $msg_count != 1;

    $error = send_mail($sender, "not authorized for ssh access", "You are not authorized to activate ssh access");
	$log_entry .= "$PID: Problem sending mail via $sendmail: Error code $error\n" if $error;
	lockLOG();
	print LOG  $log_entry or die "Cannot write to logfile: $log_file";
	unlockLOG();
}
else
{
	if ($ip_found == 1)
	{
		my $hosts_allow = "/etc/hosts.allow";
		if ($error = write_ip($hosts_allow, $ip_address))
		{
			lockLOG();
			print LOG $error or die "Cannot write to logfile: $log_file";
			unlockLOG();
			die $error;
		}
		# Beginn Firewall-Kommunikation

		# Ende Firewall-Kommunikation
		$log_entry = "$PID: *** activated $ip_address ***\n";
        $error = send_mail($sender, "Authorization for ssh access granted", "ssh access activated for $ip_address");
		$log_entry .= "$PID: Problem sending mail via $sendmail: Error code $error\n" if $error;
		lockLOG();
		print LOG $log_entry or die "Cannot write to logfile: $log_file";
		unlockLOG();

	}
	else
	{
		$log_entry = "$PID: *** error (unparseable ip address) ***\n";
        $error = send_mail($sender, "Error in ssh access activation", "could not parse your ip address");
		$log_entry .= "$PID: Problem sending mail via $sendmail: Error code $error\n" if $error;
		lockLOG();
		print LOG $log_entry or die "Cannot write to logfile: $log_file";
		unlockLOG();
	}
}
close(LOG);
exit;

##### End of main program

##### Subroutines start here

sub lockLOG ()
{
	flock(LOG, LOCK_EX) or die "Cannot lock file: $log_file";
	seek(LOG, 0, 2) or die "Cannot seek in file: $log_file";
}

sub unlockLOG ()
{
	flock(LOG, LOCK_UN)  or die "Cannot unlock file: $log_file";
}

sub write_ip ($$)
{
	my($hostsfile, $ip) = @_;

	return "$PID: Cannot open file $hostsfile" unless open(HOSTS_ALLOW, ">>$hostsfile");
	return "$PID: Cannot lock file $hostsfile" unless flock(HOSTS_ALLOW, LOCK_EX);
	return "$PID: Cannot seek file $hostsfile" unless seek(HOSTS_ALLOW, 0, 2);
	return "$PID: Cannot append $ip to $hostsfile\n" unless print HOSTS_ALLOW "sshd: $ip\n";
	return "Cannot unlock file: $hostsfile" unless flock(HOSTS_ALLOW, LOCK_UN);
	unless (close(HOSTS_ALLOW))
	{
		$error = "$PID: Cannot close $hostsfile\n";
		lockLOG();
		print LOG $error or die "Cannot write to logfile: $log_file";
		unlockLOG();
	}

	return undef;
}

sub time_difference_ok ($$$$$)
{
	my($hours, $min, $day, $mon, $year) = @_;

	my $sent_time = timelocal(0, $min, $hours, $day, $mon - 1, $year);
	my $diff = abs($sent_time - time);

	return $diff < $time_diff;
}

sub send_mail ($$$)
{
	my($sender, $subject, $body) = @_;
	return "$PID: Cannot open pipe to mail program" unless open(MAILER, "| $sendmail");

	return "$PID: Cannot write to mail program" unless print MAILER "To: $sender\n";
	return "$PID: Cannot write to mail program" unless print MAILER "Subject: $subject\n";
	return "$PID: Cannot write to mail program" unless print MAILER "\n";  # Leerzeile zum Abtrennen des Headers
	return "$PID: Cannot write to mail program" unless print MAILER "$body\n";

	unless (close(MAILER))
	{
		$error = "$PID: Cannot close pipe to mail program\n";
		lockLOG();
		print LOG $error or die "Cannot write to logfile: $log_file";
		unlockLOG();
	}

	return undef;
}
